'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var jsCore = require('@lorisleiva/js-core');
var jsWeb3jsAdapters = require('@lorisleiva/js-web3js-adapters');
var web3_js = require('@solana/web3.js');

class Web3JsRpc {
  constructor(context, endpoint, rpcOptions) {
    this.context = context;
    this.connection = new web3_js.Connection(endpoint, rpcOptions);
    this.cluster = jsCore.resolveClusterFromEndpoint(endpoint);
  }
  getEndpoint() {
    return this.connection.rpcEndpoint;
  }
  getCluster() {
    return this.cluster;
  }
  async getAccount(publicKey, options = {}) {
    const account = await this.connection.getAccountInfo(jsWeb3jsAdapters.toWeb3JsPublicKey(publicKey), options);
    return this.parseMaybeAccount(account, publicKey);
  }
  async getAccounts(publicKeys, options = {}) {
    const accounts = await this.connection.getMultipleAccountsInfo(publicKeys.map(jsWeb3jsAdapters.toWeb3JsPublicKey), options);
    return accounts.map((account, index) => this.parseMaybeAccount(account, publicKeys[index]));
  }
  async getProgramAccounts(programId, options = {}) {
    const accounts = await this.connection.getProgramAccounts(jsWeb3jsAdapters.toWeb3JsPublicKey(programId), {
      ...options,
      filters: options.filters?.map(filter => this.parseDataFilter(filter))
    });
    return accounts.map(({
      pubkey,
      account
    }) => this.parseAccount(account, jsWeb3jsAdapters.fromWeb3JsPublicKey(pubkey)));
  }
  async getBalance(publicKey, options = {}) {
    const balanceInLamports = await this.connection.getBalance(jsWeb3jsAdapters.toWeb3JsPublicKey(publicKey), options);
    return jsCore.lamports(balanceInLamports);
  }
  async getRent(bytes, options = {}) {
    const rentFor = bytes => this.connection.getMinimumBalanceForRentExemption(bytes, options.commitment);
    if (options.includesHeaderBytes ?? false) {
      const headerRent = await rentFor(0);
      const rentPerByte = BigInt(headerRent) / BigInt(jsCore.ACCOUNT_HEADER_SIZE);
      return jsCore.lamports(rentPerByte * BigInt(bytes));
    }
    return jsCore.lamports(await rentFor(bytes));
  }
  async getLatestBlockhash(options = {}) {
    return this.connection.getLatestBlockhash(options);
  }
  async getTransaction(signature, options = {}) {
    const response = await this.connection.getTransaction(jsCore.base58.deserialize(signature)[0], {
      commitment: options.commitment,
      maxSupportedTransactionVersion: 0
    });
    if (!response) {
      return null;
    }
    if (!response.meta) {
      // TODO: Custom error.
      throw new Error('Transaction meta is missing.');
    }
    const {
      transaction,
      meta
    } = response;
    const message = jsWeb3jsAdapters.fromWeb3JsMessage(transaction.message);
    const mapPublicKey = key => jsWeb3jsAdapters.fromWeb3JsPublicKey(new web3_js.PublicKey(key));
    const mapTokenBalance = tokenBalance => ({
      accountIndex: tokenBalance.accountIndex,
      amount: jsCore.createAmount(tokenBalance.uiTokenAmount.amount, 'splToken', tokenBalance.uiTokenAmount.decimals),
      mint: mapPublicKey(tokenBalance.mint),
      owner: tokenBalance.owner ? mapPublicKey(tokenBalance.owner) : null
    });
    return {
      message,
      serializedMessage: this.context.transactions.serializeMessage(message),
      signatures: transaction.signatures.map(jsCore.base58.serialize),
      meta: {
        fee: jsCore.lamports(meta.fee),
        logs: meta.logMessages ?? [],
        preBalances: meta.preBalances.map(jsCore.lamports),
        postBalances: meta.postBalances.map(jsCore.lamports),
        preTokenBalances: (meta.preTokenBalances ?? []).map(mapTokenBalance),
        postTokenBalances: (meta.postTokenBalances ?? []).map(mapTokenBalance),
        innerInstructions: meta.innerInstructions?.map(inner => ({
          index: inner.index,
          instructions: inner.instructions.map(instruction => ({
            programIndex: instruction.programIdIndex,
            accountIndexes: instruction.accounts,
            data: jsCore.base58.serialize(instruction.data)
          }))
        })) ?? null,
        loadedAddresses: {
          writable: (meta.loadedAddresses?.writable ?? []).map(jsWeb3jsAdapters.fromWeb3JsPublicKey),
          readonly: (meta.loadedAddresses?.readonly ?? []).map(jsWeb3jsAdapters.fromWeb3JsPublicKey)
        },
        computeUnitsConsumed: meta.computeUnitsConsumed ? BigInt(meta.computeUnitsConsumed) : null,
        err: meta.err
      }
    };
  }
  async accountExists(publicKey, options = {}) {
    return !jsCore.isZeroAmount(await this.getBalance(publicKey, options));
  }
  async airdrop(publicKey, amount, options = {}) {
    const signature = await this.connection.requestAirdrop(jsWeb3jsAdapters.toWeb3JsPublicKey(publicKey), Number(amount.basisPoints));
    if (options.strategy) {
      await this.confirmTransaction(jsCore.base58.serialize(signature), options);
      return;
    }
    await this.confirmTransaction(jsCore.base58.serialize(signature), {
      ...options,
      strategy: {
        type: 'blockhash',
        ...(await this.getLatestBlockhash())
      }
    });
  }
  async call(method, params, options = {}) {
    const client = this.connection._rpcClient;
    const resolvedParams = this._resolveCallParams(params ? [...params] : [], options.commitment, options.extra);
    return new Promise((resolve, reject) => {
      const callback = (error, response) => error ? reject(error) : resolve(response.result);
      if (options.id) {
        client.request(method, resolvedParams, options.id, callback);
      } else {
        client.request(method, resolvedParams, callback);
      }
    });
  }
  _resolveCallParams(args, override, extra) {
    const commitment = override || this.connection.commitment;
    if (commitment || extra) {
      let options = {};
      if (commitment) {
        options.commitment = commitment;
      }
      if (extra) {
        options = {
          ...options,
          ...extra
        };
      }
      args.push(options);
    }
    return args;
  }
  async sendTransaction(transaction, options = {}) {
    try {
      const signature = await this.connection.sendRawTransaction(this.context.transactions.serialize(transaction), options);
      return jsCore.base58.serialize(signature);
    } catch (error) {
      let resolvedError = null;
      if (error instanceof Error && 'logs' in error) {
        resolvedError = this.context.programs.resolveError(error, transaction);
      }
      throw resolvedError || error;
    }
  }
  async confirmTransaction(signature, options) {
    return this.connection.confirmTransaction(this.parseConfirmStrategy(signature, options), options.commitment);
  }
  parseAccount(account, publicKey) {
    return {
      executable: account.executable,
      owner: jsWeb3jsAdapters.fromWeb3JsPublicKey(account.owner),
      lamports: jsCore.lamports(account.lamports),
      rentEpoch: account.rentEpoch,
      publicKey,
      data: new Uint8Array(account.data)
    };
  }
  parseMaybeAccount(account, publicKey) {
    return account ? {
      ...this.parseAccount(account, publicKey),
      exists: true
    } : {
      exists: false,
      publicKey
    };
  }
  parseDataFilter(filter) {
    if (!('memcmp' in filter)) return filter;
    const {
      bytes,
      ...rest
    } = filter.memcmp;
    return {
      memcmp: {
        ...rest,
        bytes: jsCore.base58.deserialize(bytes)[0]
      }
    };
  }
  parseConfirmStrategy(signature, options) {
    if (options.strategy.type === 'blockhash') {
      return {
        ...options.strategy,
        signature: jsCore.base58.deserialize(signature)[0]
      };
    }
    return {
      ...options.strategy,
      signature: jsCore.base58.deserialize(signature)[0],
      nonceAccountPubkey: jsWeb3jsAdapters.toWeb3JsPublicKey(options.strategy.nonceAccountPubkey)
    };
  }
}

exports.Web3JsRpc = Web3JsRpc;
//# sourceMappingURL=Web3JsRpc.cjs.map
